@@ -16,7 +16,10 @@ __author__ = 'HPE'
 
 from base64 import b64decode
 import json
+import os
 import re
+import shutil
+import tempfile
 
 from OpenSSL.crypto import FILETYPE_ASN1
 from OpenSSL.crypto import load_certificate
@@ -865,6 +868,7 @@ class RedfishOperations(operations.IloOperations):
         :param ignore : A boolean param, True when Secure_boot needs to be
                ignored. If passed False, Secure_boot security param will
                not be ignored. If nothing passed default will be False.
+        :raises: IloError, on an error from iLO.
         """
         try:
             self._update_security_parameter(sec_param="Secure Boot",
@@ -877,6 +881,69 @@ class RedfishOperations(operations.IloOperations):
             LOG.debug(msg)
             raise exception.IloError(msg)
 
+    def create_csr(self, path, csr_params):
+        """Creates the Certificate Signing Request.
+
+        :param path: directory to store csr file.
+        :param csr_params: A dictionary containing all the necessary
+               information required to create CSR.
+        :raises: IloError, on an error from iLO.
+        """
+        sushy_man = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        try:
+            cert_request = (
+                sushy_man.securityservice.https_certificate_uri.generate_csr(
+                    csr_params))
+
+            dir = os.path.join(path, 'cert')
+
+            if not os.path.exists(dir):
+                os.makedirs(dir, 0o755)
+
+            (fd, temp_file) = tempfile.mkstemp(suffix='.csr')
+
+            with open(temp_file, 'w') as f:
+                f.write(cert_request)
+
+            shutil.copy(temp_file, dir)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to create the '
+                          'certificate signing request. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def add_https_certificate(self, cert_file):
+        """Adds the signed https certificate to the iLO.
+
+        :param certificate: Signed HTTPS certificate file.
+        :raises: IloError, on an error from iLO.
+        """
+        sushy_man = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        try:
+            with open(cert_file, 'r') as f:
+                data = json.dumps(f.read())
+            p = re.sub(r"\"", "", data)
+            q = re.sub(r"\\n", "\r\n", p).rstrip()
+
+            c_list = re.findall(_CERTIFICATE_PATTERN, q, re.DOTALL)
+
+            if len(c_list) == 0:
+                msg = (self._("No valid certificate in %(cert_file)s.") %
+                       {"cert_file": cert_file})
+                LOG.debug(msg)
+                raise exception.IloError(msg)
+
+            cert = c_list[0]
+            sushy_man.securityservice.https_certificate_uri.import_certificate(
+                cert)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to import the '
+                          'given certificate. '
+                          'Error: %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
     def get_security_dashboard_values(self):
         """Gets all the parameters related to security dashboard.
 
@@ -1692,10 +1759,55 @@ class RedfishOperations(operations.IloOperations):
             msg = 'TLS certificate cannot be upload in BIOS boot mode'
             raise exception.IloCommandNotSupportedInBiosError(msg)
 
-    def remove_tls_certificate(self, cert_file_list=[]):
-        """Removes the TLS certificate from the iLO.
+    def _get_fps_from_file(self, cert_file):
+        """Gets the finger prints from the certificate file.
 
-        :param cert_file_list: List of TLS certificate files
+        Parse the file passed in to get the certificates. For each certificate,
+        find the fingerprint by calculating the digest of base64 decoded
+        content. Finally return the list of the fingerprints.
+
+        :param cert_file: TLS certificate file containing one or more
+               certificates.
+
+        :return: Returns the list of FPs for the certificates in the file.
+        """
+        fp_list = []
+        with open(cert_file, 'r') as f:
+            data = json.dumps(f.read())
+            p = re.sub(r"\"", "", data)
+            q = re.sub(r"\\n", "\r\n", p).rstrip()
+
+            c_list = re.findall(_CERTIFICATE_PATTERN, q, re.DOTALL)
+
+            if len(c_list) == 0:
+                LOG.warning("Could not find any valid certificate in "
+                            "%(cert_file)s. Ignoring." %
+                            {"cert_file": cert_file})
+                return fp_list
+
+            for content in c_list:
+                pem_lines = [line.strip() for line in (
+                    content.strip().split('\n'))]
+
+                try:
+                    der_data = b64decode(''.join(pem_lines[1:-1]))
+                except ValueError:
+                    LOG.warning("Illegal base64 encountered "
+                                "in the certificate.")
+                else:
+                    cert = load_certificate(FILETYPE_ASN1, der_data)
+                    fp = cert.digest('sha256').decode('ascii')
+                    fp_list.append(fp)
+        return fp_list
+
+    def remove_tls_certificate(self, cert_file_list=[],
+                               excl_cert_file_list=[]):
+        """Removes the TLS CA certificates from the iLO.
+
+        :param cert_file_list: List of TLS CA certificate files
+        :param excl_cert_file_list: List of TLS CA certificate files to be
+               retained on the iLO. These certificates will not be
+               removed from the iLO.
 
         :raises: IloError, on an error from iLO.
         :raises: IloCommandNotSupportedError, if the command is
@@ -1709,49 +1821,42 @@ class RedfishOperations(operations.IloOperations):
 
         cert_dict = {}
         del_cert_list = []
+        exc_fp_list = []
+
+        for exc_cert_file in excl_cert_file_list:
+            efp_list = self._get_fps_from_file(exc_cert_file)
+            exc_fp_list.extend(efp_list)
+
+        LOG.debug("Excluding certificates with FingerPrints: %(exc_fp_list)s",
+                  {'exc_fp_list': exc_fp_list})
 
         if not cert_file_list:
             tls_certificates = (sushy_system.bios_settings.tls_config.
                                 tls_certificates)
             for cert in tls_certificates:
                 fp = cert.get("FingerPrint")
+
+                if fp not in exc_fp_list:
+                    cert_fp = {
+                        "FingerPrint": fp
+                    }
+                    del_cert_list.append(cert_fp)
+
+        else:
+            all_fp_list = []
+
+            for cert_file in cert_file_list:
+                afp_list = self._get_fps_from_file(cert_file)
+                all_fp_list.extend(afp_list)
+
+            final_fp_set = set(all_fp_list) - set(exc_fp_list)
+
+            for fp in final_fp_set:
                 cert_fp = {
                     "FingerPrint": fp
                 }
                 del_cert_list.append(cert_fp)
 
-        else:
-            for cert_file in cert_file_list:
-                with open(cert_file, 'r') as f:
-                    data = json.dumps(f.read())
-                    p = re.sub(r"\"", "", data)
-                    q = re.sub(r"\\n", "\r\n", p).rstrip()
-
-                    c_list = re.findall(_CERTIFICATE_PATTERN, q, re.DOTALL)
-
-                    if len(c_list) == 0:
-                        LOG.warning("Could not find any valid certificate in "
-                                    "%(cert_file)s. Ignoring." %
-                                    {"cert_file": cert_file})
-                        continue
-
-                    for content in c_list:
-                        pem_lines = [line.strip() for line in (
-                            content.strip().split('\n'))]
-
-                        try:
-                            der_data = b64decode(''.join(pem_lines[1:-1]))
-                        except ValueError:
-                            LOG.warning("Illegal base64 encountered "
-                                        "in the certificate.")
-                        else:
-                            cert = load_certificate(FILETYPE_ASN1, der_data)
-                            fp = cert.digest('sha256').decode('ascii')
-                            cert_fp = {
-                                "FingerPrint": fp
-                            }
-                            del_cert_list.append(cert_fp)
-
         if len(del_cert_list) == 0:
             msg = (self._("No valid certificate in %(cert_file_list)s.") %
                    {"cert_file_list": cert_file_list})
